{-# LANGUAGE DuplicateRecordFields #-}
{-# OPTIONS_GHC -Wall #-}

module Data.Walker where

import           Data.Hashable                     (hash)
import           Data.List
import           Data.Makefile
import           Data.Makefile.Render
import           Data.SCAD
import           Data.Shared
import qualified Data.SRDF                         as SRDF
import qualified Data.Text                         as Text
import qualified Data.URDF                         as URDF
import           Debug.Trace
import           Linear.Quaternion
import           Linear.V3
import qualified Numeric.Units.Dimensional.Prelude as U

type LegName = String
type CenterOfMass = Position
type PosOffset = Position

jointNames =
  ["alfa", "beta", "gama", "delta", "epsilon", "zeta", "eta", "theta"]

data SCADModel = SCADModel {
        module_name :: SCADName,
        args        :: [Arg],
        imports     :: [Filename]
    }


data Part = Part {
            model            :: SCADModel,
            collision_models :: [SCADModel],
            part_weight      :: Maybe Mass,
            center_of_mass   :: Maybe CenterOfMass
        };

data JointConfig = JointConfig {
            min_angle     :: Angle,
            default_angle :: Angle,
            max_angle     :: Angle,
            vel_lim       :: AngularVelocity,
            effor_lim     :: Effort,
            g_steps       :: Integer

}

data Link = Link {
            link_name    :: String,
            offset       :: Pose,
            part         :: Part,
            axis         :: Axis,
            joint_config ::JointConfig
        };

data Tip = Tip {
            point :: Position
        };

data Leg = Leg {
            leg_name :: LegName,
            chain    :: [Link],
            tip      :: Tip,
            pose     :: Pose
        };

data Body = Body {
        body_part :: Part
    };

data Walker = Walker {
            walker_name :: String,
            legs        :: [Leg],
            body        :: Body
    };

getLinkName :: Leg -> String -> String
getLinkName leg name = "leg/" ++ leg_name leg ++ "/" ++ name

getStlName :: String -> SCADModel -> String
getStlName link_name model =
  "src/stl/" ++ link_name ++ "/" ++ module_name model ++ ".stl"

getURDFStlName :: String -> SCADModel -> String
getURDFStlName link_name model =
  "package://schpin_koke/" ++ getStlName link_name model

getScadName :: String -> SCADModel -> String
getScadName link_name model =
  "src/scad/gen/" ++ link_name ++ "/" ++ module_name model ++ ".scad"

getJointName = getLinkName

poseToURDFOrigin :: Pose -> URDF.Origin
poseToURDFOrigin pose = URDF.Origin (Just $ pos pose)
                                    (Just $ orientation $ rpy)
 where
  rpy = fmap (U.*~ U.radian) $ quatToRPY $ (ori pose)
  orientation (V3 r p y) = URDF.Orientation r p y


getLegJoints :: Leg -> [URDF.Joint]
getLegJoints leg =
  [first_joint] ++ map linkToURDFJoint link_triples ++ [last_joint]
 where
  link_names   = map (getLinkName leg . link_name) $ chain leg
  link_triples = take (length $ chain leg)
    $ zip3 (getLinkName leg "base" : link_names) (chain leg) (tail jointNames)

  jConfigToURDFLimit (JointConfig min_angle _ max_angle vel_lim effor_lim _) =
    URDF.Limit effor_lim vel_lim (Just min_angle) (Just max_angle)

  first_joint = URDF.Joint
    { name              = getJointName leg (head jointNames)
    , jtype             = URDF.Fixed
    , parent            = "body"
    , child             = getLinkName leg "base"
    , origin            = Just $ poseToURDFOrigin $ pose leg
    , axis              = Nothing
    , calibration       = Nothing
    , mimic             = Nothing
    , safety_controller = Nothing
    }

  linkToURDFJoint :: (String, Link, String) -> URDF.Joint
  linkToURDFJoint (p_link_name, ch_link, name) = URDF.Joint
    { name              = getJointName leg name
    , jtype = URDF.Revolute (jConfigToURDFLimit $ joint_config ch_link) Nothing
    , parent            = p_link_name
    , child             = getLinkName leg $ link_name ch_link
    , origin            = Just $ poseToURDFOrigin $ offset ch_link
    , axis              = Just $ axis ch_link
    , calibration       = Nothing
    , mimic             = Nothing
    , safety_controller = Nothing
    }
  last_joint = URDF.Joint
    { name = getJointName leg (jointNames !! (2 + length (chain leg)))
    , jtype             = URDF.Fixed
    , parent            = last link_names
    , child             = getLinkName leg "tip"
    , origin            = Just $ URDF.Origin (Just $ point $ tip leg) Nothing
    , axis              = Nothing
    , calibration       = Nothing
    , mimic             = Nothing
    , safety_controller = Nothing
    }


getPartVisual :: String -> Part -> URDF.Visual
getPartVisual link_name part = URDF.Visual
  { v_geometry = URDF.Mesh (getURDFStlName link_name $ model part)
                           (Just $ V3 0.001 0.001 0.001)
  , name       = Nothing
  , origin     = Nothing
  , material   = Just $ URDF.Material link_name (Just color) Nothing
  }
 where
  color = URDF.Color
    { r = fromIntegral (hash (link_name ++ "red") `mod` 1024) / 1024.0
    , g = fromIntegral (hash (link_name ++ "green") `mod` 1024) / 1024.0
    , b = fromIntegral (hash (link_name ++ "blue") `mod` 1024) / 1024.0
    , a = 1
    }

getPartCollisions :: String -> Part -> [URDF.Collision]
getPartCollisions link_name part = map f (collision_models $ part)
 where
  f model = URDF.Collision
    { geometry = URDF.Mesh (getURDFStlName link_name model)
                           (Just $ V3 0.001 0.001 0.001)
    , name     = Nothing
    , origin   = Nothing
    }


getLegLinks :: Leg -> [URDF.Link]
getLegLinks leg =
  [URDF.Link (getLinkName leg "base") [] Nothing []]
    ++ map linkToURDFLink (chain leg)
    ++ [URDF.Link (getLinkName leg "tip") [] Nothing []]
 where
  linkToURDFLink leg_link = URDF.Link
    { name      = get_name leg_link
    , visuals   = [getPartVisual (get_name leg_link) $ part leg_link]
    , inertial  = Nothing
    , collision = getPartCollisions (get_name leg_link) $ part leg_link
    }
  get_name leg_link = getLinkName leg $ link_name leg_link


getBodyLink :: Body -> URDF.Link
getBodyLink body = URDF.Link
  { name      = "body"
  , visuals   = [getPartVisual "body" $ body_part body]
  , inertial  = Nothing
  , collision = getPartCollisions "body" $ body_part body
  }


getURDF :: Walker -> URDF.URDF
getURDF walker = URDF.URDF (walker_name walker) joints links
 where
  joints = concatMap getLegJoints (legs walker)
  links  = getBodyLink (body walker) : (concatMap getLegLinks $ legs walker)


getLegSRDFChain :: Leg -> SRDF.Chain
getLegSRDFChain leg = SRDF.Chain (link_name $ head urdf_chain)
                                 (link_name $ last urdf_chain)
 where
  urdf_chain = getLegLinks leg
  link_name :: URDF.Link -> UrdfName
  link_name (URDF.Link name _ _ _) = name

getLegSRDFJoints :: Leg -> [SRDF.Joint]
getLegSRDFJoints leg = zipWith convert_chain (chain leg) (tail jointNames)
 where
  convert_chain (Link _ _ _ _ joint_config) joint_name =
    SRDF.Joint (getJointName leg joint_name) (build_joint_config joint_config)
  build_joint_config (JointConfig _ _ _ _ _ g_steps) =
    [SRDF.GenericNode "graph_steps" [SRDF.Attribute "value" "16"] []]

getSRDF :: Walker -> SRDF.SRDF
getSRDF walker = SRDF.SRDF (walker_name walker)
                           [leg_group]
                           items
                           [schpin_config]
 where
  leg_group = SRDF.Group "legs"
                         []
                         (concatMap getLegSRDFJoints $ legs walker)
                         (map getLegSRDFChain $ legs walker)
                         []
                         []
  items         = []
  schpin_config = SRDF.GenericNode
    "schpin_config"
    []
    [ SRDF.GenericNode "environment"  [SRDF.Attribute "resolution" "0.05"] []
    , SRDF.GenericNode "leg_workarea" [SRDF.Attribute "resolution" "0.01"] []
    , SRDF.GenericNode "leg_mapping"  [SRDF.Attribute "resolution" "0.01"] []
    ]

iterateAllParts :: Walker -> (String -> Part -> a) -> [a]
iterateAllParts walker f =
  (f "body" $ body_part $ body walker) : (concatMap proc_legs $ legs walker)
 where
  proc_legs leg =
    map (\x -> f (getLinkName leg $ link_name x) $ part x) $ chain leg

getSTLRenderList :: Walker -> [(Filename, Filename)]
getSTLRenderList walker = concat $ iterateAllParts walker part_to_files
 where
  model_to_files name model = (getScadName name model, getStlName name model)

  part_to_files :: String -> Part -> [(Filename, Filename)]
  part_to_files name part =
    (model_to_files name $ model part)
      : (map (model_to_files name) $ collision_models part)

getMakefileRenderer :: Walker -> Makefile
getMakefileRenderer walker = Makefile
  { entries = ( Rule (Target $ Text.pack "all")
                     (map list_to_deb $ getSTLRenderList walker)
                     []
              )
    : (map list_to_rule $ getSTLRenderList walker)
  }
 where
  list_to_rule :: (Filename, Filename) -> Entry
  list_to_rule (scad_file, stl_file) = Rule
    (Target $ Text.pack stl_file)
    [Dependency $ Text.pack scad_file]
    (stlRenderCommands scad_file stl_file)
  list_to_deb (_, stl_file) = Dependency $ Text.pack stl_file

renderWalkerStl :: Walker -> IO ()
renderWalkerStl walker = do
  writeMakefile "Makefile" $ getMakefileRenderer walker

getSCADModels :: Walker -> [(Filename, SCADModel)]
getSCADModels walkie = concat $ iterateAllParts walkie part_to_models
 where
  part_to_models :: String -> Part -> [(Filename, SCADModel)]
  part_to_models name part =
    map (\x -> (getScadName name x, x)) $ (model part) : (collision_models part)

modelToSCAD :: (Filename, SCADModel) -> (Filename, [SCAD])
modelToSCAD (file, model) = (file, s)
 where
  s =
    map to_import_f (imports model) ++ [Module (module_name model) (args model)]
  to_import_f :: Filename -> SCAD
  to_import_f filename =
    Use
      $  concat (replicate (length (filter (== '/') file) - 2) "../")
      ++ filename

poseToSCAD :: Pose -> [SCAD] -> SCAD
poseToSCAD pose child = Translate (pos pose) [Rotate (ori pose) child]

getSCADSkeleton :: String -> Walker -> (Part -> SCAD) -> (Filename, [SCAD])
getSCADSkeleton filename (Walker _ legs body) part_to_skeleton =
  ( filename
  , [Use "../../body.scad"]
    ++ map Use         (concatMap get_imports legs)
    ++ map to_skeleton legs
    ++ [part_to_skeleton $ body_part body]
  )
 where
  get_imports (Leg _ chain _ _) =
    map (\x -> "../../" ++ x) $ concatMap (imports . model . part) chain

  to_skeleton :: Leg -> SCAD
  to_skeleton (Leg _ chain tip pose) = poseToSCAD pose $ link_to_skeleton chain

  link_to_skeleton :: [Link] -> [SCAD]
  link_to_skeleton [] = []
  link_to_skeleton (link:rest) =
    [ poseToSCAD
        (offset link)
        [ Rotate (toQuat (axis link) (default_angle $ joint_config link))
                 (part_to_skeleton (part link) : link_to_skeleton rest)
        ]
    ]

partToVisual (Part model _ _ _) = Module (module_name model) (args model)
partToCollision (Part _ collision_models _ _) =
  Debug $ map (\x -> Module (module_name x) (args x)) collision_models


getSCADFiles :: Walker -> [(Filename, [SCAD])]
getSCADFiles walkie = map modelToSCAD (getSCADModels walkie)

getSCADSkeletons :: String -> Walker -> [(Filename, [SCAD])]
getSCADSkeletons name walkie =
  [gen_f (name ++ ".scad") partToVisual]
    ++ [gen_f (name ++ "_col.scad") partToCollision]
 where
  gen_f name f = getSCADSkeleton ("src/scad/gen/skeletons/" ++ name) walkie f
